#
# Copyright (C) 2009, 2012 UNINETT AS
#
# This file is part of Network Administration Visualized (NAV).
#
# NAV is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.  You should have received a copy of the GNU General Public License
# along with NAV. If not, see <http://www.gnu.org/licenses/>.
#

import os
from datetime import datetime
import logging

from django.db.models import Q

from nav.web import webfrontConfig
from nav.config import read_flat_config
from nav.models.msgmaint import Message
from nav.models.event import AlertHistory
from nav.models.manage import Netbox

_logger = logging.getLogger('nav.web.tools.utils')

class Tool(object):
    """Class representing a tool"""
    def __init__(self, name, uri, icon, description, priority=0, display=True):
        self.name = name
        self.uri = uri
        self.icon = icon
        self.description = description
        self.priority = int(priority)
        self.display = display

    def __unicode__(self):
        return self.name

    def __eq__(self, other):
        return self.name == other.name

    def __cmp__(self, other):
        return cmp(self.priority, other.priority)

    def __repr__(self):
        return "Tool('%s')" % self.name


def quick_read(filename):
    """Read and return the contents of a file, or None if something went wrong.
    """
    try:
        return file(filename).read().strip()
    except IOError:
        return None

def current_messages():
    """Finds current messages"""
    return Message.objects.filter(
        publish_start__lt=datetime.today().isoformat(' '),
        publish_end__gt=datetime.today().isoformat(' '),
        replaced_by__isnull=True
    )

def boxes_down():
    """Finds boxes that are down and not currently on maintenance"""
    infinity = datetime.max
    on_maintenance = Netbox.objects.filter(
        alerthistory__event_type='maintenanceState',
        alerthistory__end_time__gte=infinity,
    )
    boxes_down = AlertHistory.objects.select_related(
        'netbox'
    ).filter(
        Q(netbox__up=Netbox.UP_DOWN) | Q(netbox__up=Netbox.UP_SHADOW),
        end_time__gte=infinity,
        event_type='boxState'
    ).exclude(netbox__in=on_maintenance).order_by('-start_time')
    return boxes_down

def tool_list(account):
    """Get the list of tools existing in the tool directories"""

    def parse_tool(tool_file):
        """Parse the tool file and return a Tool object"""
        lines = open(tool_file).read().splitlines()
        return Tool(**dict(
            [[y.strip() for y in x.split('=')] for x in lines if x]))

    paths = {}
    if webfrontConfig.has_option('toolbox', 'path'):
        paths = webfrontConfig.get('toolbox', 'path').split(os.pathsep)
    else:
        return None
    
    tool_list = []
    for path in paths:
        if os.access(path, os.F_OK):
            filelist = os.listdir(path)
            for filename in filelist:
                if filename[-5:] == '.tool':
                    fullpath = os.path.join(path, filename)
                    try:
                        tool = parse_tool(fullpath)
                    except Exception, error:
                        _logger.error('Error parsing tool in %s: %s',
                                      filename, error)
                        continue

                    if account.has_perm('web_access', tool.uri):
                        tool_list.append(tool)
    tool_list.sort()
    return tool_list
